/*
 * PyroIgnitionHardwareBasic.c
 *
 * Created: 08.11.2016 23:00:18
 *  Author: Deneriel

 * 2017.01.06	Code bereinigt
  
INPUTs:			
(PullUp on)			
Address0					PD5
Address1					PD6
Address2					PD7
Address3					PB0

OUTPUTs:
Channel1					PC3
Channel2					PC2
Channel3					PC1
Channel4					PC0
Channel5					PB4
Channel6					PB3
Channel7					PB2
Channel8					PB1

UART:
RX							PD0
TX							PD1
DriverEnable|~ReceiverEnable	PD2
  
Vorbereitet:
Address4					PD4
Strommessung				PC4 (ADC/TWI)
unused						PC5 (ADC/TWI)
unused						PD3


UART Protokoll
-----------------
Protokoll bernommen von PyroIgnitionControl, geschrieben von PyroBla

http://software.fx-firingsystem.de/
http://www.feuerwerk-forum.de/showthread.php?t=24096

UART-Settings: 9600,8n1

Paketlaenge: 4 Byte
1: Startbyte, immer 0xff
2: Adresse, 1-30
3: Kanal, 1-99
4: CRC-8 aus Byte 2+3

Ein Paket braucht immer ca. 8ms. Nach Empfang wird der Puffer sofort geleert.
Der AVR ist schnell genug um das Paket immer zu verarbeiten, whrend das nchste empfangen wird. Es braucht
daher nur einen einfachen Empfangspuffer.


*/ 

#define F_CPU 7372800
#define BAUD 9600
#include <util/setbaud.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/crc16.h>

#define ADDRESS		0
#define CHANNEL		1
#define CHECKSUM	2

struct {
	uint8_t data[3];
	uint8_t position:2;
	uint8_t complete:1;
} _rxbuffer;
uint8_t _commandbuffer[3];

uint8_t _deviceaddress;
uint8_t crc = 0;

#define MAXCHANNEL		8
#define IGNITEVALUE		7200;

volatile uint16_t _ignitiontimer[MAXCHANNEL+1];
volatile uint8_t firepin[MAXCHANNEL+1] = {0x00, PC3, PC2, PC1, PC0, PB4, PB3, PB2, PB1 };
volatile uint8_t *fireport[MAXCHANNEL+1] = {0x00, &PORTC, &PORTC, &PORTC, &PORTC, &PORTB, &PORTB, &PORTB, &PORTB };
	
void init()
{
	

	DDRB = (1<<PB4) | (1<<PB3) | (1<<PB2) | (1<<PB1);
	DDRC = (1<<PC3) | (1<<PC2) | (1<<PC1) | (1<<PC0);
	DDRD = (1<<PD2);
	
	PORTB = (1<<PB0);
	PORTC = 0;
	PORTD = (1<<PD7) | (1<<PD6) | (1<<PD5) | (1<<PD4);			// PD4 = optionales 5. Adressbit
	
	// read Device Address from I/Os
	_deviceaddress = (   ( ((PIND>>5) & 0x07) | ((PINB & 0x01)<<3) | (PIND & 0x10))  ^ 0x1f) +1;
	
	// UART 9600,8n1, RX Int enable
	// RS485 DriverEnable = 0 / ReceiverEnable = 1 -> PD2 = 0; wird noch nicht genutztt.
	UCSRB |= (1<<RXCIE) | (1<<RXEN) | (1<<TXEN);
	UCSRC |= (1<<URSEL) | (0<<UCSZ2) | (1<<UCSZ1) | (1<<UCSZ0);
	
	UBRRH = UBRRH_VALUE;
	UBRRL = UBRRL_VALUE;
	
	// UART Puffer vorbereiten
	_rxbuffer.complete = 0;
	_rxbuffer.position = 0;
	
	// Timer 0 Init	
	// Prescaler clk/256, Overflow Interrupt enable
	TCCR0 |= (1<<CS01) | (0<<CS01) | (0<<CS00);		// Prescaler = clk/256 
	TIMSK |= (1<<TOIE0);							
	
}

void fire(uint8_t chn, uint8_t status)
{
	if(status>0)
	{
		// Status > 0 -> znden, Ausgang = high
		// Status == 0 -> Zndung vorbei, Ausgang = low
		*fireport[chn] |= (1<<firepin[chn]);
		_ignitiontimer[chn] = IGNITEVALUE;
	} else {
		*fireport[chn] &= ~(1<<firepin[chn]);
		_ignitiontimer[chn] = 0;
	}
}


ISR(USART_RXC_vect)
{
	cli();
	uint8_t rcv;
    rcv = UDR;
	if(rcv == 255)
	{
		// 255 ist als Startbyte reserviert, daher Buffer leeren.
		_rxbuffer.position = 0;
		_rxbuffer.complete = 0;
	} else {
			
	    if(_rxbuffer.complete == 0)		// bei vollem Buffer Empfangene Bytes verwerfen
		{
			_rxbuffer.data[_rxbuffer.position] = rcv;

			if(_rxbuffer.position < 2)		// Buffer voll?
			{
				_rxbuffer.position++;
			} else {			
				_rxbuffer.complete = 1;
			}			
		}		
	} 
	sei();
}

ISR(TIMER0_OVF_vect)
{
	cli();
	for(uint8_t i=0;i<=MAXCHANNEL;i++)
	{
		if(_ignitiontimer[i] > 0)
		{
			_ignitiontimer[i]--;
		} else {
			fire(i, 0);
		}
	}
	sei();
}

int main(void)
{
    init();
	sei();
	
	while(1)
    {	
							
        if(_rxbuffer.complete)
		{
			// UART Puffer leeren
			cli();
			_commandbuffer[ADDRESS] = _rxbuffer.data[0];
			_commandbuffer[CHANNEL] = _rxbuffer.data[1];
			_commandbuffer[CHECKSUM] = _rxbuffer.data[2];
			_rxbuffer.complete = 0;		// Puffer zurcksetzen damit das Paket nicht mehrfach ausgewertet wird.
			_rxbuffer.position = 0;				
			sei();
			
			if(_commandbuffer[ADDRESS] == _deviceaddress)
			{
				// Datenpaket ist fr uns.
				// CRC-8 Checksumme prfen, wenn ok Befehl ausfhren.
				crc = 0x00;
				crc = _crc_ibutton_update(crc, _commandbuffer[ADDRESS]);
				crc = _crc_ibutton_update(crc, _commandbuffer[CHANNEL]);
				if (crc == _commandbuffer[CHECKSUM])
				{
					if((_commandbuffer[CHANNEL] >0) && (_commandbuffer[CHANNEL] <= MAXCHANNEL))
					{
						fire(_commandbuffer[CHANNEL], 1);
					}
				}			
			} 
		}	
    }
}